"""
FNO code borrowed from (https://arxiv.org/pdf/2010.08895.pdf)
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
import matplotlib.pyplot as plt

import operator
import sys
from functools import reduce
from functools import partial
from timeit import default_timer
from utilities3 import *
from decimal import Decimal

from Adam import Adam

torch.manual_seed(0)
np.random.seed(0)


################################################################
#  1d fourier layer
################################################################
class SpectralConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, modes1):
        super(SpectralConv1d, self).__init__()

        """
        1D Fourier layer. It does FFT, linear transform, and Inverse FFT.    
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1  #Number of Fourier modes to multiply, at most floor(N/2) + 1

        self.scale = (1 / (in_channels*out_channels))
        self.weights1 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, self.modes1, dtype=torch.cfloat))

    # Complex multiplication
    def compl_mul1d(self, input, weights):
        # (batch, in_channel, x ), (in_channel, out_channel, x) -> (batch, out_channel, x)
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.out_channels, x.size(-1)//2 + 1,  device=x.device, dtype=torch.cfloat)
        out_ft[:, :, :self.modes1] = self.compl_mul1d(x_ft[:, :, :self.modes1], self.weights1)

        #Return to physical space
        x = torch.fft.irfft(out_ft, n=x.size(-1))
        return x
    
    
################################################################
#  1d kernel correction
################################################################
class KernelCorrection1d(nn.Module):
    def __init__(self, layer):
        super(KernelCorrection1d, self).__init__()

        """
        1D Corrected Layer. It modifies the kernel to enforce a Dirichlet boundary condition.   
        """

        self.layer = layer
        
            
    def forward(self, x , bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right):
        """
        Forward layer to enforce the boundary as well as the correction terms for
            a general operator K
            
        We apply T^{1}_Neumann, K and T^{2}_{Neumann} on x
        """
        
        e1 = torch.zeros_like(x)
        e1[:, :, 0] = 1
        K00 = self.layer(e1)[0,0,0]
    
            
        if torch.abs(K00) < 10**(-7):
            K00 = 10**(-7)


        Kx = self.layer(x)
        xnew1 = x.clone()
        
        xnew1[:, :, 0] = 2*xnew1[:, :, 0] - Kx[:, :, 0]/K00   
        xnew1 = self.layer(xnew1.clone())
        
        xnew1[:, :, 0] = x[:, :, 0]
        
                    
        
        return xnew1
        
        

class FNO1d(nn.Module):
    def __init__(self, modes, width):
        super(FNO1d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .
        
        input: the solution of the initial condition and location (a(x), x)
        input shape: (batchsize, x=s, c=2)
        output: the solution of a later timestep
        output shape: (batchsize, x=s, c=1)
        """

        self.modes1 = modes
        self.width = width
        self.padding = 2 # pad the domain if input is non-periodic
        self.fc0 = nn.Linear(2, self.width) # input channel is 2: (a(x), x)

        self.conv0 = SpectralConv1d(self.width, self.width, self.modes1)
        self.conv1 = SpectralConv1d(self.width, self.width, self.modes1)
        self.conv2 = SpectralConv1d(self.width, self.width, self.modes1)
        self.conv3 = SpectralConv1d(self.width, self.width, self.modes1)
        
        self.kernel_correction0 = KernelCorrection1d(self.conv0)
        self.kernel_correction1 = KernelCorrection1d(self.conv1)
        self.kernel_correction2 = KernelCorrection1d(self.conv2)
        self.kernel_correction3 = KernelCorrection1d(self.conv3)
        
        self.w0 = nn.Conv1d(self.width, self.width, 1)
        self.w1 = nn.Conv1d(self.width, self.width, 1)
        self.w2 = nn.Conv1d(self.width, self.width, 1)
        self.w3 = nn.Conv1d(self.width, self.width, 1)
        
        self.w0_correction = KernelCorrection1d(self.w0)
        self.w1_correction = KernelCorrection1d(self.w1)
        self.w2_correction = KernelCorrection1d(self.w2)
        self.w3_correction = KernelCorrection1d(self.w3)
        
        
        self.fc1 = nn.Linear(self.width, 128)
        self.fc2 = nn.Linear(128, 1)

                

    def forward(self, x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right):
        """
        Forward function of the Neural Operator.
        
        Args:
            x: array representing the input of the Neural Operator, 
                given by the initial condition of the PDE
            w: array representing the exact boundary condition on the right bdy
        """
        
        
        grid = self.get_grid(x.shape, x.device)
        
        x = torch.cat((x, grid), dim=-1)     
        x = self.fc0(x)
        
        x = x.permute(0, 2, 1)
                
        # First corrected fourier layer
        x1 = self.kernel_correction0(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right) 
        x2 = self.w0_correction(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x = (x1 + x2)/2  
        x[:, :, 1:x.shape[2]-1] = F.gelu(x[:, :, 1:x.shape[2]-1].clone())

        
        # Second corrected Fourier layer
        x1 = self.kernel_correction1(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x2 = self.w1_correction(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x = (x1 + x2)/2
        x[:, :, 1:x.shape[2]-1] = F.gelu(x[:, :, 1:x.shape[2]-1].clone())
        
        
        
        # Third corrected Fourier layer
        x1 = self.kernel_correction2(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x2 = self.w2_correction(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x = (x1 + x2)/2
        x[:, :, 1:x.shape[2]-1] = F.gelu(x[:, :, 1:x.shape[2]-1].clone())
        
        # Fourth corrected Fourier layer

        x1 = self.kernel_correction3(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x2 = self.w3_correction(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)
        x = (x1 + x2)/2
        x = x.permute(0, 2, 1)
        
        x = self.fc1(x)
        x[:, 1:x.shape[1]-1, :] = F.gelu(x[:, 1:x.shape[1]-1, :].clone())
        
        x = self.fc2(x)
        
                
        
        x[:, -1] = (bdy_right - coeffs_finite_difference_right[-2]*x[:, -2] - coeffs_finite_difference_right[-3]*x[:, -3] - coeffs_finite_difference_right[-4]*x[:, -4])/coeffs_finite_difference_right[-1]
        
        x[:, 0] = x[:, 1]
        

        

        return x

    def get_grid(self, shape, device):
        batchsize, size_x = shape[0], shape[1]
        gridx = torch.tensor(np.linspace(0, 1, size_x), dtype=torch.float)
        gridx = gridx.reshape(1, size_x, 1).repeat([batchsize, 1, 1])
        return gridx.to(device)

################################################################
#  configurations
################################################################
ntrain = 500
ntest = 100

sub = int(sys.argv[1]) #subsampling rate
# sub = 1 #subsampling rate
# N = 500 // sub #total grid size divided by the subsampling rate
N = 4096 // sub #total grid size divided by the subsampling rate

s = N

print(s)

batch_size = 20
learning_rate = 0.001

epochs = 500
step_size = 50
gamma = 0.5

modes = 16
width = 64


################################################################
# read data
################################################################



x_data = np.genfromtxt('initial_heat_1D.csv', delimiter=',')[:,::sub]
x_data = x_data.astype(np.float32)
x_data = torch.from_numpy(x_data)
print(x_data.shape)


y_data = np.genfromtxt('sol_heat_1D.csv', delimiter=',')[:,::sub]
y_data = y_data.astype(np.float32)
y_data = torch.from_numpy(y_data)


x_train = x_data[:ntrain,:]
y_train = y_data[:ntrain,:]
x_test = x_data[-ntest:,:]
y_test = y_data[-ntest:,:]


x_train = x_train.reshape(ntrain,s,1)
x_test = x_test.reshape(ntest,s,1)


train_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_train, y_train), batch_size=batch_size, shuffle=True)
test_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_test, y_test), batch_size=batch_size, shuffle=False)

# model
model = FNO1d(modes, width).cuda()
print("Number of parameters in the model {}".format(count_params(model)))

################################################################
# training and evaluation
################################################################
optimizer = Adam(model.parameters(), lr=learning_rate, weight_decay=1e-4)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)

myloss = LpLoss(size_average=False)
absLoss = LpLoss(size_average=False)

err_bdy_vec_near_bdy = np.zeros(epochs)
err_bdy_vec_test = np.zeros(epochs)

h = 1/(N-1)

coeffs_finite_difference_left = np.array([-11/(6*h), 3/h, -3/(2*h), 1/(3*h)])
coeffs_finite_difference_left = coeffs_finite_difference_left.astype(np.float32)
coeffs_finite_difference_left = torch.from_numpy(coeffs_finite_difference_left).cuda()

coeffs_finite_difference_right = np.array([-1/(3*h), 3/(2*h), -3/h , 11/(6*h)])
coeffs_finite_difference_right = coeffs_finite_difference_right.astype(np.float32)
coeffs_finite_difference_right = torch.from_numpy(coeffs_finite_difference_right).cuda()

neumann_bdy_left = 0.0

U = np.array([5])
U = torch.from_numpy(U).cuda()

t = np.array([0.5])
t = torch.from_numpy(t).cuda()
neumann_bdy_right = U*torch.sin(torch.pi*t)

for ep in range(epochs):
    model.train()
    t1 = default_timer()
    train_mse = 0
    train_l2 = 0
    for x, y in train_loader:
        x, y = x.cuda(), y.cuda()
        bdy_left = neumann_bdy_left
#         bdy_left = y[:, 0]
        bdy_right = neumann_bdy_right
        
        optimizer.zero_grad()
        out = model(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)

        mse = F.mse_loss(out.view(batch_size, -1), y.view(batch_size, -1), reduction='mean')
        l2 = myloss(out.view(batch_size, -1), y.view(batch_size, -1))
        l2.backward() # use the l2 relative loss

        optimizer.step()
        train_mse += mse.item()
        train_l2 += l2.item()
        

    scheduler.step()
    model.eval()
    test_l2 = 0.0
    test_l2_near_bdy = 0.0
    with torch.no_grad():
        for x, y in test_loader:
            x, y = x.cuda(), y.cuda()
            bdy_left = neumann_bdy_left
#             bdy_left = y[:, 0]
            bdy_right = neumann_bdy_right

            out = model(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right)


            test_l2 += myloss(out.view(batch_size, -1), y.view(batch_size, -1)).item()

    train_mse /= len(train_loader)
    train_l2 /= ntrain
    test_l2 /= ntest
    err_bdy_vec_near_bdy[ep] = test_l2_near_bdy
    err_bdy_vec_test[ep] = absLoss.abs(out.view(batch_size, -1), y.view(batch_size, -1))

    t2 = default_timer()
    print(ep, t2-t1, train_mse, train_l2, test_l2)

# x_trial = torch.eye(s)
# kernel_values = torch.zeros((s,s))
# for i in range(s):
#     kernel_values[:,i] = model(x_trial[:, i].reshape(1, s, 1).cuda(), w, bdy_left).reshape(s)
    
pred = torch.zeros(y_test.shape)
index = 0
test_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_test, y_test), batch_size=1, shuffle=False)
test_loss_average = 0
test_l2_absolute = 0
with torch.no_grad():
    for x, y in test_loader:
        test_l2 = 0
        x, y = x.cuda(), y.cuda()
        bdy_left = neumann_bdy_left
#         bdy_left = y[:, 0]
        bdy_right = neumann_bdy_right


        out = model(x, bdy_left, bdy_right, coeffs_finite_difference_left, coeffs_finite_difference_right).view(-1)
        pred[index] = out

        test_l2 += myloss(out.view(1, -1), y.view(1, -1)).item()
        test_l2_absolute += torch.norm(y[0,:]-out)
        test_loss_average += test_l2
        print(index, test_l2)
        index = index + 1
    print("")
    print("Relative test error is {}".format(test_loss_average/ntest))
    print("")
    print("Absolute test error is {}".format(test_l2_absolute/ntest))

