"""
FNO code borrowed from (https://arxiv.org/pdf/2010.08895.pdf)
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

import matplotlib.pyplot as plt

import operator
import sys
from functools import reduce
from functools import partial

from timeit import default_timer
from utilities3 import *

from Adam import Adam

torch.manual_seed(0)
np.random.seed(0)


################################################################
# fourier layer
################################################################
class SpectralConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, modes1, modes2):
        super(SpectralConv2d, self).__init__()

        """
        2D Fourier layer. It does FFT, linear transform, and Inverse FFT.    
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1 #Number of Fourier modes to multiply, at most floor(N/2) + 1
        self.modes2 = modes2

        self.scale = (1 / (in_channels * out_channels))
        self.weights1 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, self.modes1, self.modes2, dtype=torch.cfloat))
        self.weights2 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, self.modes1, self.modes2, dtype=torch.cfloat))

    # Complex multiplication
    def compl_mul2d(self, input, weights):
        # (batch, in_channel, x,y ), (in_channel, out_channel, x,y) -> (batch, out_channel, x,y)
        return torch.einsum("bixy,ioxy->boxy", input, weights)

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft2(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.out_channels,  x.size(-2), x.size(-1)//2 + 1, dtype=torch.cfloat, device=x.device)
        out_ft[:, :, :self.modes1, :self.modes2] = \
            self.compl_mul2d(x_ft[:, :, :self.modes1, :self.modes2], self.weights1)
        out_ft[:, :, -self.modes1:, :self.modes2] = \
            self.compl_mul2d(x_ft[:, :, -self.modes1:, :self.modes2], self.weights2)

        #Return to physical space
        x = torch.fft.irfft2(out_ft, s=(x.size(-2), x.size(-1)))
        return x

class FNO2d(nn.Module):
    def __init__(self, modes1, modes2,  width):
        super(FNO2d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .
        
        input: the solution of the coefficient function and locations (a(x, y), x, y)
        input shape: (batchsize, x=s, y=s, c=3)
        output: the solution 
        output shape: (batchsize, x=s, y=s, c=1)
        """

        self.modes1 = modes1
        self.modes2 = modes2
        self.width = width
        self.padding = 9 # pad the domain if input is non-periodic
        self.fc0 = nn.Linear(3, self.width) # input channel is 3: (a(x, y), x, y)

        self.conv0 = SpectralConv2d(self.width, self.width, self.modes1, self.modes2)
        self.conv1 = SpectralConv2d(self.width, self.width, self.modes1, self.modes2)
        self.conv2 = SpectralConv2d(self.width, self.width, self.modes1, self.modes2)
        self.conv3 = SpectralConv2d(self.width, self.width, self.modes1, self.modes2)
        self.w0 = nn.Conv2d(self.width, self.width, 1)
        self.w1 = nn.Conv2d(self.width, self.width, 1)
        self.w2 = nn.Conv2d(self.width, self.width, 1)
        self.w3 = nn.Conv2d(self.width, self.width, 1)

        self.fc1 = nn.Linear(self.width, 128)
        self.fc2 = nn.Linear(128, 1)

    def forward(self, x):
        grid = self.get_grid(x.shape, x.device)
        x = torch.cat((x, grid), dim=-1)
        x = self.fc0(x)
        x = x.permute(0, 3, 1, 2)
#         x = F.pad(x, [0,self.padding, 0,self.padding])

        alpha = 0.5
        beta = 0.5
        
        x1 = self.conv0(x)
        x2 = self.w0(x)
        x = x1 + x2
        x[:,:,0, :] = alpha*x[:,:,0, :] + beta* x[:,:,-1, :]
        x[:,:,-1, :] = x[:,:,0, :]
        x = F.gelu(x)

        x1 = self.conv1(x)
        x2 = self.w1(x)
        x = x1 + x2
        x[:,:,0, :] = alpha*x[:,:,0, :] + beta* x[:,:,-1, :]
        x[:,:,-1, :] = x[:,:,0, :]
        x = F.gelu(x)

        x1 = self.conv2(x)
        x2 = self.w2(x)
        x = x1 + x2
        x[:,:,0, :] = alpha*x[:,:,0, :] + beta* x[:,:,-1, :]
        x[:,:,-1, :] = x[:,:,0, :]
        x = F.gelu(x)

        x1 = self.conv3(x)
        x2 = self.w3(x)
        x = x1 + x2
        x[:,:,0, :] = alpha*x[:,:,0, :] + beta* x[:,:,-1, :]
        x[:,:,-1, :] = x[:,:,0, :]

#         x = x[..., :-self.padding, :-self.padding]
        x = x.permute(0, 2, 3, 1)
        x = self.fc1(x)
        x[:,0,:, :] = alpha*x[:,0,:, :] + beta* x[:,-1,:, :]
        x[:,-1,:, :] = x[:,0,:, :]
        x = F.gelu(x)
        x = self.fc2(x)
        
        x[:,0,:, :] = alpha*x[:,0,:, :] + beta* x[:,-1,:, :]
        x[:,-1,:, :] = x[:,0,:, :]
        
        return x
    
    def get_grid(self, shape, device):
        batchsize, size_x, size_y = shape[0], shape[1], shape[2]
        gridx = torch.tensor(np.linspace(0, 1, size_x), dtype=torch.float)
        gridx = gridx.reshape(1, size_x, 1, 1).repeat([batchsize, 1, size_y, 1])
        gridy = torch.tensor(np.linspace(0, 1, size_y), dtype=torch.float)
        gridy = gridy.reshape(1, 1, size_y, 1).repeat([batchsize, size_x, 1, 1])
        return torch.cat((gridx, gridy), dim=-1).to(device)

################################################################
# configs
################################################################
# TRAIN_PATH = 'piececonst_r421_N1024_smooth1.mat'
# TEST_PATH = 'piececonst_r421_N1024_smooth2.mat'

ntrain = 1000
ntest = 200

resolution_data = 500

batch_size = 20
learning_rate = 0.001

epochs = 500
step_size = 100
gamma = 0.5

modes = 12
width = 32

sub = int(sys.argv[1]) #subsampling rate
N = 128 // sub #total grid size divided by the subsampling rate
# N = 500 // sub #total grid size divided by the subsampling rate
s = N
T_in = 1

T = 25

# r = 5
# # h = int(((421 - 1)/r) + 1)
# h = int(((241 - 1)/r) + 1)
# s = h

################################################################
# load data and data normalization
################################################################

# reader = MatReader(TRAIN_PATH)
# x_train = reader.read_field('coeff')[:ntrain,::r,::r][:,:s,:s]
# y_train = reader.read_field('sol')[:ntrain,::r,::r][:,:s,:s]

# reader.load_file(TEST_PATH)
# x_test = reader.read_field('coeff')[:ntest,::r,::r][:,:s,:s]
# y_test = reader.read_field('sol')[:ntest,::r,::r][:,:s,:s]

# data_all_training = scipy.io.loadmat('piececonst_r241_N1024_smooth1.mat')

# data_all = scipy.io.loadmat('stokes_2D_data_with_time.mat')
data_all = scipy.io.loadmat('burgers_pino.mat')
# data_all = scipy.io.loadmat('burgers_Dirichlet_2D_data_with_time.mat')
# data_all = scipy.io.loadmat('heat_2D_data_with_time.mat')


# x_train = data_all['a'][:ntrain,::sub,:T_in]
x_train = data_all['input'][:ntrain,::sub]
x_train = x_train.reshape(ntrain,s,1)
# x_train = np.swapaxes(x_train, 1,2)
x_train = x_train.astype(np.float32)
x_train = torch.from_numpy(x_train)

# y_train = data_all['u'][:ntrain,::sub,T_in:T+T_in]
# y_train = data_all['u'][:ntrain,::sub,-T:]
y_train = data_all['output'][:ntrain,-T:, ::sub]
y_train = np.swapaxes(y_train, 1,2)
y_train = y_train.astype(np.float32)
y_train = torch.from_numpy(y_train)

# x_test = data_all['a'][-ntest:,::sub,:T_in]
x_test = data_all['input'][-ntest:,::sub]
x_test = x_test.reshape(ntest,s,1)
# x_test = np.swapaxes(x_test, 1,2)
x_test = x_test.astype(np.float32)
x_test = torch.from_numpy(x_test)

# y_test = data_all['u'][-ntest:,::sub,-T:]
y_test = data_all['output'][-ntest:,-T:, ::sub]
y_test = np.swapaxes(y_test, 1,2)
y_test = y_test.astype(np.float32)
y_test = torch.from_numpy(y_test)



# x_normalizer = UnitGaussianNormalizer(x_train)
# x_train = x_normalizer.encode(x_train)
# x_test = x_normalizer.encode(x_test)

# y_normalizer = UnitGaussianNormalizer(y_train)
# y_train = y_normalizer.encode(y_train)

x_train = x_train.reshape(ntrain,s,1,T_in).repeat([1,1,T,1])
print(x_train.shape)
x_test = x_test.reshape(ntest,s,1,T_in).repeat([1,1,T,1])

alpha = 0.5
beta = 0.5

y_train[:, 0, :] = alpha*y_train[:, 0, :] + beta*y_train[:, -1, :]
y_train[:, -1, :] = y_train[:, 0, :]

y_test[:, 0, :] = alpha*y_test[:, 0, :] + beta*y_test[:, -1, :]
y_test[:, -1, :] = y_test[:, 0, :]

# y_test[:, -1, :] = y_test[:, 0, :]

train_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_train, y_train), batch_size=batch_size, shuffle=True)
test_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_test, y_test), batch_size=batch_size, shuffle=False)

################################################################
# training and evaluation
################################################################
model = FNO2d(modes, modes, width).cuda()
print(count_params(model))

optimizer = Adam(model.parameters(), lr=learning_rate, weight_decay=1e-4)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)

myloss = LpLoss(size_average=False)
# y_normalizer.cuda()
for ep in range(epochs):
    model.train()
    t1 = default_timer()
    train_l2 = 0
    for x, y in train_loader:
        x, y = x.cuda(), y.cuda()

        optimizer.zero_grad()
        out = model(x).reshape(batch_size, s, T)
#         print(out.shape)
#         print(y.shape)
#         out = y_normalizer.decode(out)
#         y = y_normalizer.decode(y)

        loss = myloss(out.view(batch_size,-1), y.view(batch_size,-1))
        loss.backward()

        optimizer.step()
        train_l2 += loss.item()

    scheduler.step()

    model.eval()
    test_l2 = 0.0
    with torch.no_grad():
        for x, y in test_loader:
            x, y = x.cuda(), y.cuda()

            out = model(x).reshape(batch_size, s, T)
#             out = y_normalizer.decode(out)

            test_l2 += myloss(out.view(batch_size,-1), y.view(batch_size,-1)).item()

    train_l2/= ntrain
    test_l2 /= ntest

    t2 = default_timer()
    print(ep, t2-t1, train_l2, test_l2)
    
pred = torch.zeros(y_test.shape)

test_loss_average = 0.0
test_loss_Dirichlet = 0.0
test_loss_periodic = 0.0

index = 0
test_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(x_test, y_test), batch_size=1, shuffle=False)
with torch.no_grad():
    for x, y in test_loader:
        test_l2 = 0
        x, y = x.cuda(), y.cuda()
        out = model(x).reshape(1, s, T)
#         out = y_normalizer.decode(out)
#         out = model(x)[0][:,:,0]
#         pred[index] = out

        test_l2 += myloss(out.view(1, -1), y.view(1, -1)).item()
        test_loss_Dirichlet += torch.norm(out[0, -1, :] - y[0, -1, :]) + torch.norm(out[0, 0, :] - y[0, 0, :])
        test_loss_periodic += torch.norm(out[0, -1, :] - out[0, 0, :])
        test_loss_average += test_l2
        print(index, test_l2)
        index = index + 1
        
    np.savetxt('results_stokes_nu_0_point_1.csv', np.array([test_loss_average/ntest, (test_loss_Dirichlet/ntest).cpu().detach().numpy()]))
    print("Test error is {}".format(test_loss_average/ntest))
    print("Error at Dirichlet bdy is {}".format(test_loss_Dirichlet/ntest))
    print("Error periodic bdy is {}".format(test_loss_periodic/ntest))

# with torch.no_grad():
#     for x, y in test_loader:
#         test_l2 = 0
#         x, y = x.cuda(), y.cuda()

#         out = model(x).view(-1).reshape(s, s)
#         pred[index] = out

#         test_l2 += myloss(out.view(1, -1), y.view(1, -1)).item()
#         print(index, test_l2)
#         index = index + 1
